
#include <ball_tracking/tracker.hpp>
#include <ball_tracking/img_proc.hpp>
#include <ball_tracking/utils.hpp>
#include <algorithm>
#include <json.hpp>

using namespace cv;
using namespace std;
using json = nlohmann::json;

namespace ball_tracking {

  /*
   * Private functions
   */
  namespace {

    class GaussSmooth {
      private:
        int size;
        double sigma;
      public:
        GaussSmooth(const json& conf) {
          size = conf.at("size");
          sigma = conf.at("sigma");
        }

        cv::Mat operator()(cv::InputArray _src) {
          Mat dst;
          GaussianBlur(_src, dst, Size(size,size), sigma, sigma);
          return dst;
        }
    };

    /**
     * Color and background based logistic regression
     */
    class CB_log_reg {
      private:
        Mat bkg; //!< Background image
        Mat weights; //!< Model weights
        vector<preproc> pre; //!< Preprocessing functions for the images
      public:
 
        Mat pre_chain(cv::InputArray src) {
          Mat tmp = src.getMat();
          for (auto p : pre) {
            tmp = p(tmp);
          }
          return tmp;
        }
 
        cv::Mat operator()(cv::InputArray _src) {
          Mat src = pre_chain(_src);
          if (bkg.empty()) {
            bkg = src;
          }
          return quadf_log_reg(src, bkg, weights);
        }

        CB_log_reg(const json& conf) {
          weights = json2cvmat(conf.at("weights"));
          if (conf.count("gauss_smooth")) {
            pre.push_back(GaussSmooth(conf.at("gauss_smooth")));
          }
          if (conf.count("background")) {
            bkg = pre_chain( imread(conf.at("background"), CV_LOAD_IMAGE_COLOR) );
            if (!bkg.data) {
              throw std::logic_error("File not found for background image");
            }
          }
        }
    };

    /**
     * OpenCV Blob detection algorithm
     * TODO: Continue
     */
    class CV_blob_detect {
      private:
        double thresh;
        SimpleBlobDetector::Params params;
        Ptr<SimpleBlobDetector> detector;
      public:
        void set_blob_params(const json& conf) {
          if (conf.count("filterByArea")) {
            params.filterByArea = conf.at("filterByArea");
            params.minArea = conf.at("minArea");
            params.maxArea = conf.at("maxArea");
          }
          if (conf.count("filterByCircularity")) {
            params.filterByCircularity = conf.at("filterByCircularity");
            params.minCircularity = conf.at("minCircularity");
            params.maxCircularity = conf.at("maxCircularity");
          }
        }

        CV_blob_detect(const json& conf) {
          set_blob_params(conf);
#if CV_MAJOR_VERSION < 3
          detector = Ptr<SimpleBlobDetector>(new SimpleBlobDetector(params));
#else
          detector = SimpleBlobDetector::create(params);
#endif
        }

        vector<KeyPoint> operator()(cv::InputArray bin_img) {
          vector<KeyPoint> ans;
          detector->detect(bin_img, ans);
          return ans;
        }
    };

  };

  class Binarizer::Impl {
    public:
      double thresh;

      Impl(const json& conf) {
        if (conf.count("p_thresh")) {
          double p = conf.at("p_thresh");
          thresh = log(p/(1-p)); //logit function
        } else if (conf.count("thresh")) {
          thresh = conf.at(thresh);
        }
      }

      Mat bin(cv::InputArray src) {
        Mat bin_img;
        threshold(src, bin_img, thresh, 1.0, THRESH_BINARY_INV);
        bin_img.convertTo(bin_img, CV_8U, 255);
        return bin_img;
      }
  };

  Binarizer::Binarizer(const nlohmann::json& params) {
    _impl = unique_ptr<Impl>(new Impl(params));
  }

  Binarizer::~Binarizer() = default;

  cv::Mat Binarizer::operator()(cv::InputArray src) {
    return _impl->bin(src);
  }


  class BallLogLikelihood::Impl {
    public:
      preproc bll;

      Impl(const json& config) {
        const string& type = config.at("type");
        const json& conf = config.at("conf");
        if (type == "cb_log_reg") {
          bll = CB_log_reg(conf);
        } else {
          throw std::logic_error("Type of ball log likelihood algorithm not recognized");
        }
      }
  };

  BallLogLikelihood::BallLogLikelihood(const nlohmann::json& params) {
    _impl = unique_ptr<Impl>(new Impl(params));
  }
      
  BallLogLikelihood::~BallLogLikelihood() = default;

  cv::Mat BallLogLikelihood::operator()(cv::InputArray src) {
    return _impl->bll(src);
  }


  class FindBallBlob::Impl {
    public:
      blob_finder bf;

      Impl(const json& config) {
        const string& type = config.at("type");
        const json& conf = config.at("conf");
        if (type == "cv_blob_detect") {
          bf = CV_blob_detect(conf);
        } else {
          throw std::logic_error("Type of the blob finder algorithm not recognized");
        }
      }
  };

  FindBallBlob::FindBallBlob(const nlohmann::json& params) {
    _impl = unique_ptr<Impl>(new Impl(params));
  }

  FindBallBlob::~FindBallBlob() = default;

  std::vector<cv::KeyPoint> FindBallBlob::operator()(cv::InputArray src) {
    return _impl->bf(src);
  }


};
